	output "rom.bin"

;----------------------------------------------
;----------------------------------------------
;  variables in unused rom (ram during startup)
;----------------------------------------------
;----------------------------------------------

varbas:		equ $386E

fatbas:		equ varbas+30
storage_begin:	equ fatbas
fat1_begin:	equ storage_begin+4
root_cluster:	equ fat1_begin+4
last_cluster:	equ root_cluster+4
filelen:	equ last_cluster+4
filecluster:	equ filelen+4
l2spc:		equ filecluster+4
spc:		equ l2spc+1
filesector:	equ spc+1
bufix:		equ filesector+1
bufmax:		equ bufix+2
fatbuf:		equ bufmax+2

romstack:	equ $3D00

;--------------------------------------
;--------------------------------------
;  CODE
;--------------------------------------
;--------------------------------------
	ld	sp,$8000	; Stack at the end of page #0
	
	ld	a,1		; Blue border
	out	($FE),a
	ld	a,$0E		; Yellow over Blue attributes
	call	cls
	
	; -------------- Load ZX ROM from SPI Flash -------------------
	ld	a,$1F		; MOSI, SCK y /SS Low
	out	($FF),a
	ld	a,3		; Command Read
	call	spibyte
	ld	a,127		; MSB
	call	spibyte
	ld	a,0
	call	spibyte
	ld	a,0
	call	spibyte		; LSB
	ld	hl,0
lbuc:	call	spibyte
	ld	(hl),a
	inc	hl
	ld	a,$40
	cp	h
	jr	nz,lbuc
	ld	a,$3F		; /SS high
	out	($FF),a

	ld	sp,romstack	; All RAM free

;--------------------------------------------------
;--------------------------------------------------
;	SD Card init
;--------------------------------------------------
;--------------------------------------------------
	
	call	sd_init
	or	a
	jp	nz,gorom
	call	FAT_init
	or	a
	jp	nz,gorom
	
;--------------------------------------------------
;--------------------------------------------------
;	If an SD card is found select File to load
;--------------------------------------------------
;--------------------------------------------------

	jp	filesel

;--------------------------------------------------
;--------------------------------------------------
;	Load SNA file
;--------------------------------------------------
;--------------------------------------------------
loadSNA:
	ld	hl,txto2
	call	puts

	ld	ix,fatbuf
	call	rdfile
	ld	de,varbas
	ld	hl,fatbuf
	ld	bc,27
	ldir
	ld	de,$4000
	ld	bc,512-27
	ldir
	
blsna:	ld	ix,fatbuf
	call	rdfile
	ld	a,b
	ld	hl,fatbuf
	ldir
	cp	2
	jr	z,blsna

	jp	gosna

;--------------------------------------------------
;--------------------------------------------------
;	BOOT ROM BASIC
;--------------------------------------------------
;--------------------------------------------------
	
gorom:	ld	hl,fatbas	; $3900 to $3cff all filled with $ff
	ld	de,fatbas+1	; needed for IM2 programs like HeadOverHeels
	ld	bc,$3d00-fatbas-1
	ld	a,$ff
	ld	(hl),a
	ldir
	
	ld	sp,0
	ld	hl,txto3
	call	puts
br1:	in	a,($BF)
	and	4
	jr	z,br1	
	
	ld	hl,boot0	; Copy trampoline code to RAM
	ld	de,$4000
	ld	bc,endboot-bootSNA1
	ldir
	jp	$4000		; and execute it

;--------------------------------------------------
;--------------------------------------------------
;	BOOT  SNA
;--------------------------------------------------
;--------------------------------------------------

gosna:	ld	hl,fatbas	; $3900 to $3cff all filled with $ff
	ld	de,fatbas+1
	ld	bc,$3d00-fatbas-1
	ld	a,$ff
	ld	(hl),a
	ldir

	; copy appropriate trampoline depending on interrupts
	ld	hl,bootSNA0
	ld	de,fatbuf
	ld	bc,endboot-bootSNA1
	ld	a,(varbas+19)	; Interrupt ON if !=0
	or	a
	jr	z,bsna3
	ld	hl,bootSNA1
bsna3:	ldir
				; Z80 register values
	ld	a,(varbas+0)
	ld	i,a
	exx
	ld	hl,(varbas+1)
	ld	de,(varbas+3)
	ld	bc,(varbas+5)
	exx
	ex	af,af'
	ld	hl,(varbas+7)
	push	hl
	pop	af
	ex	af,af'
	ld	a,(varbas+26)	; Border color
	out	($FE),a
	
	im	1
	ld	a,(varbas+25)	; IM ?
	cp	2
	jr	nz,imod2
	im	2
imod2:	
	ld	hl,(varbas+21)	; AF
	push	hl
	pop	af
	ld	hl,(varbas+9)
	ld	de,(varbas+11)
	ld	ix,(varbas+15)
	ld	iy,(varbas+17)
	ld	sp,(varbas+23)
	jp	fatbuf

boot0:		; boot ROM
	ld	bc,$37FF	; BOOT off, pag 11
	out	(c),b
	jp	0
bootSNA0:	; boot SNA, IRQ OFF	
	ld	bc,$37FF	; BOOT off, pag 11
	out	(c),b
	ld	bc,(varbas+13)
	ret
bootSNA1:	; boot SNA, IRQ ON	
	ld	bc,$37FF	; BOOT off, pag 11
	out	(c),b
	ld	bc,(varbas+13)
	ei
	ret
endboot:	

; ---------- SPI Transfer -------------
;
; Parameters:	A: byte to send
; Returns:	A: gyte receided
; Modiffies:	nothing else

spibyte:
	push	bc
	ld	b,8
	ld	c,a
spb1:	in	a,($FF)
	sla	a
	sla	c
	rra
	and	$BF
	out	($FF),a
	or	$40
	out	($FF),a
	in	a,($FF)
	and	1
	jr	z,spb2
	inc	c
spb2:	djnz	spb1
	in	a,($FF)
	and	$BF
	out	($FF),a
	ld	a,c
	pop	bc
	ret

;--------------------------------------
; SD card selection
;
; Parameters:	none
; Returns:	none
; Modiffies:	none

SD_CS_L:
	push	af
	in 	a,($FF)
	and	$EF
	out	($FF),a
rtcsl:	pop	af
	ret

SD_CS_H:
	push	af
	in 	a,($FF)
	or	$10
	out	($FF),a
	jr	rtcsl

;--------------------------------------
; SD command
;
; Parameters:	A: command
;		DE:HL argument (32 bits)
; Returns:	A: R1 state
; Modiffies:	nothing else

sd_cmd:
	call	SD_CS_L
	push	bc
	ld	b,$87
	or	a
	jr	nz,sdcm1
	ld	b,$95
sdcm1:	or	$40
	call	spibyte	; CMD
	ld	a,d
	call	spibyte	; arg MSB
	ld	a,e
	call	spibyte
	ld	a,h
	call	spibyte
	ld	a,l
	call	spibyte	; arg LSB
	ld	a,b
	call	spibyte	; CRC

	ld	b,0
sdcm2:	ld	a,$FF	; wait response
	call	spibyte
	cp	$FF
	jr	nz,sdcm3
	djnz	sdcm2

sdcm3:	ld	b,a
	ld	a,$ff
	call	spibyte	
	call	SD_CS_H
	ld	a,b
	pop	bc
	ret
;--------------------------------------
; SD command, 32 bit result
;
; Parameters:	A: command
;		DE:HL argument (32 bits)
; Returns:	A: R1 state
;		DE:HL 32 bit result (sometimes)
; Modiffies:	nothing else

sd_cmd32:
	call	SD_CS_L
	push	bc
	ld	b,$87
	or	a
	jr	nz,sdc31
	ld	b,$95
sdc31:	or	$40
	call	spibyte	; CMD
	ld	a,d
	call	spibyte	; arg MSB
	ld	a,e
	call	spibyte
	ld	a,h
	call	spibyte
	ld	a,l
	call	spibyte	; arg LSB
	ld	a,b
	call	spibyte	; CRC

	ld	b,0
sdc32:	ld	a,$FF	; wait response
	call	spibyte
	cp	$FF
	jr	nz,sdc33
	djnz	sdc32

sdc33:	ld	b,a
	ld	a,$ff
	call	spibyte	
	ld	d,a
	ld	a,$ff
	call	spibyte	
	ld	e,a
	ld	a,$ff
	call	spibyte	
	ld	h,a
	ld	a,$ff
	call	spibyte	
	ld	l,a
	ld	a,$ff
	call	spibyte	
	call	SD_CS_H
	ld	a,b
	pop	bc
	ret

;--------------------------------------
; SD init
;
; Parameters:	none
; Returns:	A: 0 OK
; Modiffies:	lots of registers

sd_init:
	call	SD_CS_H
	ld	b,10
sdini1:	call	spibyte
	djnz	sdini1
	call	SD_CS_L
	ld	b,32
sdini2:	xor	a	; CMD go idle state
	ld	de,0
	ld	hl,0
	call	sd_cmd
	cp	1
	jr	z,sdini3
	djnz	sdini2
	ld	a,1	; Error:timeout
	ret
sdini3: ld	b,0
sdini35:
	ld	a,8	; SEND_IF_COND
	ld	de,0
	ld	hl,$01aa
	call	sd_cmd32
	cp	1
	jr	z,sdini40
	djnz	sdini35
sdini40:
	ld	b,0
sdini4:	ld	a,55	; APP_CMD
	ld	de,0
	ld	hl,0
	call	sd_cmd
	ld	a,41	;SD_SEND_OP_COND,0x40000000
	ld	de,$4000
	call	sd_cmd	
	cp	0
	jr	z,sdini5
	djnz	sdini4
	ld	a,2	; Error card initialization failed
	ret
sdini5:	ld	b,0
sdini6:	ld	a,58	; Read OCR
	ld	de,0
	ld	hl,0
	call	sd_cmd32
	cp	0
	jr	z,sdini9
	djnz	sdini6
	ld	a,3
	ret
sdini9:	ld	a,0	;OK
	ret
	
;--------------------------------------
; SD read block
;
; Parameters:	DE:HL: nº sector
;		IX: puntero
; Returns:	A: 0 OK, !=0 Error
; Modiffies:	IX = IX+512

sd_rd_block:
	push	bc
	ld	a,17	; READ SINGLE BLOCK
	call	sd_cmd
	or	a
	jr	nz,sdrd9
	call	SD_CS_L
	ld	b,0
sdrd2:	ld	a,$ff
	call	spibyte
	cp	$fe
	jr	z,sdrd3
	djnz 	sdrd2
	ld	a,4
	jr	sdrd9
sdrd3:	ld	bc,512
sdrd4:	ld	a,$ff
	call	spibyte
	ld	(ix+0),a
	inc	ix
	dec	bc
	ld	a,b
	or	c
	jr	nz,sdrd4
	
	ld	a,$ff
	call	spibyte
	ld	a,$ff
	call	spibyte
	ld	a,$ff
	call	spibyte
	
	ld	a,0
sdrd9:	call	SD_CS_H
	pop	bc
	ret

;--------------------------------------
;--------------------------------------
; FAT init
; Parameters:	none
; Returns:	A: 0 OK, !=0 Error
; Modiffies:	lots

FAT_init:
	ld	iy,fatbas
	ld	ix,fatbuf
	ld	de,0
	ld	hl,0
	call	sd_rd_block	; tabla de particiones
	cp	0
	ret	nz
	ld	de,-2
	add	ix,de	
	ld	a,$55
	cp	(ix+0)
	ret	nz
	ld	a,$AA
	cp	(ix+1)
	ret	nz
	ld	de,-56
	add	ix,de
	ld	d,(ix+3)
	ld	e,(ix+2)
	ld	h,(ix+1)
	ld	l,(ix+0)
	
	ld	ix,fatbuf
	call	sd_rd_block	; superblock
	cp	0
	ret	nz
	
	ld	ix,fatbuf
	ld	b,(ix+$d)	; Sectors/Cluster
	ld	a,b
	ld	(spc),a
	ld	a,0
fini1:	inc	a
	srl	b
	jr	nz,fini1
	dec	a
	ld	(l2spc),a	; log2(sector_per_cluster)

	ld	b,(ix+$f)	; BC=nº de sectores reservados
	ld	c,(ix+$e)
	add	hl,bc
	jr	nc,fini2
	inc	e
	jr	nc,fini2
	inc	d
fini2:	ld	(iy+fat1_begin-fatbas),l
	ld	(iy+fat1_begin-fatbas+1),h
	ld	(iy+fat1_begin-fatbas+2),e
	ld	(iy+fat1_begin-fatbas+3),d

	ld	b,(ix+$10)	; nº de FATs
fini3:	ld	a,(ix+$24)	; ($24-$27): sectores/FAT
	add	l
	ld	l,a
	ld	a,(ix+$25)	
	adc	h
	ld	h,a
	ld	a,(ix+$26)	
	adc	e
	ld	e,a
	ld	a,(ix+$27)	
	adc	d
	ld	e,a
	djnz	fini3
	ld	(iy+storage_begin-fatbas),l	; storage begin
	ld	(iy+storage_begin-fatbas+1),h
	ld	(iy+storage_begin-fatbas+2),e
	ld	(iy+storage_begin-fatbas+3),d
	
	ld	l,(ix+$2c)	; root_cluster
	ld	h,(ix+$2d)
	ld	e,(ix+$2e)
	ld	d,(ix+$2f)
	ld	(iy+root_cluster-fatbas),l
	ld	(iy+root_cluster-fatbas+1),h
	ld	(iy+root_cluster-fatbas+2),e
	ld	(iy+root_cluster-fatbas+3),d
	
	ld	a,$ff		; last_cluster
	ld	(iy+last_cluster-fatbas),a
	ld	(iy+last_cluster-fatbas+1),a
	ld	(iy+last_cluster-fatbas+2),a
	ld	(iy+last_cluster-fatbas+3),a

	ld	a,0
	ret
;--------------------------------------
;--------------------------------------
; next_cluster
; Parameters:	DE:HL : Cluster actual
; Returns:	DE:HL : siguiente cluster
; Modiffies:	A

next_cluster:
	push	bc
	push	ix
	push	iy
	ld	iy,fatbas
	ld	a,$7F
	and	l
	ld	c,a	; c=cluster&127
	ld	a,l
	xor	(iy+last_cluster-fatbas)
	and	$80
	jr	nz,nxc1
	ld	a,h
	cp	(iy+last_cluster-fatbas+1)
	jr	nz,nxc1
	ld	a,e
	cp	(iy+last_cluster-fatbas+2)
	jr	nz,nxc1
	ld	a,d
	cp	(iy+last_cluster-fatbas+3)
	jr	z,nxc2
nxc1:	ld	(iy+last_cluster-fatbas),l	; last_cluster=cluster
	ld	(iy+last_cluster-fatbas+1),h
	ld	(iy+last_cluster-fatbas+2),e
	ld	(iy+last_cluster-fatbas+3),d
	;cluster>>7
	rl	l
	rl	h
	ld	l,h
	rl	e
	ld	h,e
	rl	d
	ld	e,d
	ld	a,0
	rl	a
	ld	d,a
	; +fat_1_begin
	ld	a,l
	add	(iy+fat1_begin-fatbas)
	ld	l,a
	ld	a,h
	add	(iy+fat1_begin-fatbas+1)
	ld	h,a
	ld	a,e
	add	(iy+fat1_begin-fatbas+2)
	ld	e,a
	ld	a,d
	add	(iy+fat1_begin-fatbas+3)
	ld	d,a
	ld	ix,fatbuf
	call	sd_rd_block
nxc2:	ld	b,0	; BC*4
	sla	c
	rl	b
	sla	c
	rl	b
	ld	ix,fatbuf
	add	ix,bc
	ld	l,(ix+0)
	ld	h,(ix+1)
	ld	e,(ix+2)
	ld	d,(ix+3)
	pop	iy
	pop	ix
	pop	bc
	ret
;--------------------------------------
;--------------------------------------
; rdfile
; Parameters:	IX : buffer
; Returns:	BC : nº bytes leidos
;		IX = IX+nbytes
; Modiffies:	A

rdfile:	push	de
	push	hl
	push	iy
	ld	iy,fatbas
	ld	a,(iy+filesector-fatbas)	; A,C = ns = filesector+1
	inc	a
	ld	c,a
	ld	l,(iy+filecluster-fatbas)	; DE:HL = nc =filecluster
	ld	h,(iy+filecluster-fatbas+1)
	ld	e,(iy+filecluster-fatbas+2)
	ld	d,(iy+filecluster-fatbas+3)
	cp	(iy+spc-fatbas)			; if (ns==SecPerCluster)
	jr	nz,rdf1
	ld	c,0				; ns = 0
	call	next_cluster
rdf1:	ld	a,(iy+filecluster-fatbas+3)	; if (filecluster>=0x0FffffF0)
	or	$F0
	and	(iy+filecluster-fatbas+2)
	and	(iy+filecluster-fatbas+1)
	cp	$FF
	jr	nz,rdf15
	ld	a,$0F
	or	(iy+filecluster-fatbas+0)
	cp	$FF
	jr	nz,rdf15
rdf91:	ld	bc,0				; return 0
	jp	rdf99
rdf15:	ld	a,fatbuf&$ff			; if (IX==fatbuf)
	cp	ixl
	jr	nz,rdf2
	ld	a,fatbuf/256
	cp	ixh
	jr	nz,rdf2
	ld	a,$ff
	ld	(iy+last_cluster-fatbas),a	; lastcluster=0xFFFFFFFF
	ld	(iy+last_cluster-fatbas+1),a
	ld	(iy+last_cluster-fatbas+2),a
	ld	(iy+last_cluster-fatbas+3),a
rdf2:	push	de	; sector=storage_begin+((filecluster-2)<<l2spc)+filesector
	push 	hl
	ld	a,(iy+filecluster-fatbas)
	sub	2
	ld	l,a
	ld	a,(iy+filecluster-fatbas+1)
	sbc	0
	ld	h,a
	ld	a,(iy+filecluster-fatbas+2)
	sbc	0
	ld	e,a
	ld	a,(iy+filecluster-fatbas+3)
	sbc	0
	ld	d,a
	ld	b,(iy+l2spc-fatbas)
rdf3:	sla	l
	rl	h
	rl	e
	rl	d
	djnz	rdf3
	ld	a,l
	add	(iy+storage_begin-fatbas)	
	ld	l,a
	ld	a,h
	adc	(iy+storage_begin-fatbas+1)
	ld	h,a
	ld	a,e
	adc	(iy+storage_begin-fatbas+2)
	ld	e,a
	ld	a,d
	adc	(iy+storage_begin-fatbas+3)
	ld	d,a
	ld	a,(iy+filesector-fatbas)
	add	l
	ld	l,a
	ld	a,0
	adc	h
	ld	h,a
	ld	a,0
	adc	e
	ld	e,a
	ld	a,0
	adc	d
	ld	d,a
	call	sd_rd_block
	pop	hl		; DE:HL = nc again
	pop	de
	cp	0
	jp	nz,rdf91
	ld	(iy+filesector-fatbas),c	; filesector=ns
	ld	(iy+filecluster-fatbas),l	; filecluster=nc
	ld	(iy+filecluster-fatbas+1),h
	ld	(iy+filecluster-fatbas+2),e
	ld	(iy+filecluster-fatbas+3),d
	
	ld	bc,512				; if (filelen>=512) bc=512
	ld	a,$FE
	and	(iy+filelen-fatbas+1)
	or	(iy+filelen-fatbas+2)
	or	(iy+filelen-fatbas+3)
	jr	nz,rdf4
	ld	c,(iy+filelen-fatbas)		; else BC=filelen
	ld	b,(iy+filelen-fatbas+1)
rdf4:	ld	a,(iy+filelen-fatbas)		; filelen-=BC
	sub	c
	ld	(iy+filelen-fatbas),a
	ld	a,(iy+filelen-fatbas+1)
	sbc	b
	ld	(iy+filelen-fatbas+1),a
	ld	a,(iy+filelen-fatbas+2)
	sbc	0
	ld	(iy+filelen-fatbas+2),a
	ld	a,(iy+filelen-fatbas+3)
	sbc	0
	ld	(iy+filelen-fatbas+3),a
	ld	a,0
rdf99:	pop	iy			; return BC
	pop	hl
	pop	de
	ret

;--------------------------------------
;--------------------------------------
; fstream
; Parameters:	none
; Returns:	A = next byte from file
; Modiffies:	IX

fstream:
	push	de
	push	hl
	ld	de,(bufix)
	ld	hl,(bufmax)
	
	ld	a,d
	xor	h
	jr	nz,fst1
	ld	a,e
	xor	l
	jr	nz,fst1
	push	bc
	ld	ix,fatbuf
	call	rdfile
	ld	(bufmax),bc
	pop	bc
	ld	de,0
	ld	(bufix),de
fst1:	ld	hl,fatbuf
	add	hl,de
	ld	a,(hl)
	inc	de
	ld	(bufix),de
	
	pop	hl
	pop	de
	ret

;--------------------------------------
;--------------------------------------
; UART routines
;--------------------------------------
;--------------------------------------

puts:	ld	a,(hl)
	or	a
	ret	z
	call	putch
	inc	hl
	jr	puts

putch:	push	af
pch1:	in	a,($BF)
	and	2
	jr	z,pch1
	pop	af
	out	($7F),a
	ret
	
getch:	in	a,($BF)
	and	1
	jr	z,getch
	in	a,($7F)
	ret

;--------------------------------------	
; Print A in hex

prthex:	push	bc
	ld	c,a
	srl	a
	srl	a
	srl	a
	srl	a
	add	48
	cp	58
	jr	c,phx1
	add	7
phx1:	call	putch
	ld	a,15
	and	c
	add	48
	cp	58
	jr	c,phx2
	add	7
phx2:	call	putch
	pop	bc
	ret
;--------------------------------------	
; HEX dump: 256 bytes
; arg: HL: start address
;

hdump2:	push	bc
	push	hl
	ld	c,16
	jr	hdx1
;--------------------------------------	
; HEX dump: fatbuf
; 512 bytes
	
hdump:	push	bc
	push	hl
	ld	hl,fatbuf
	ld	c,32
hdx1:	ld	b,16
hdx2:	ld	a,(hl)
	inc	hl
	call	prthex
	ld	a,32
	call	putch
	djnz	hdx2
	
	ld	b,16
	ld	a,l
	sub	b
	ld	l,a
	ld	a,h
	sbc	0
	ld	h,a
	
hdx3:	ld	a,(hl)
	inc	hl
	cp	32
	jr	nc,hdx4
	ld	a,'.'
hdx4:	or	a
	jp	ns,hdx5
	ld	a,'.'
hdx5:	call	putch
	djnz	hdx3
	
	ld	a,10
	call	putch
	dec	c
	jr	nz,hdx1
	pop	hl
	pop	bc
	ret
;--------------------------------------
;--------------------------------------
; VIDEO routines
;--------------------------------------
;--------------------------------------
;	print character Acc at DE address
vputch:	push	hl
	push	de
	push	bc
	sub	32
	ld	l,a
	ld	h,0
	add	hl,hl
	add	hl,hl
	add	hl,hl
	ld	bc,$3D00	; ZX FONT in ROM
	add	hl,bc
	ld	b,8
vpc1:	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	d
	djnz	vpc1
	pop	bc
	pop	de
	pop	hl
	ret

;	Print string pointed by HL at DE address
vputs:	ld	a,(hl)
	or	a
	ret	z
	call	vputch
	inc	hl
	inc	de
	jr	vputs

txto1:	db "ZX-FPGA boot ROM.  J. Arias 2022",0
txto4:	db "SD-HC, FAT32, boot File SELECTOR",0
txto5:	db "Q:up, A:down, O:left, P:right,  "
	db "     ENTER:select   B:basic     ",0 
txto2:	db "booting SNA\r\n",0
txto3:	db "booting ROM\r\n",0
txtdir:	db " D",0
txtfil:	db "  ",0
txtotp:	db " TAP File preloaded to High RAM "
	db "  Push F1 to play after LOAD \"\"",0

;----------------------------------------------------------
;----------------------------------------------------------
;	Interactive File Selection
;----------------------------------------------------------
;----------------------------------------------------------

; variables

fselbas:	equ	$4000+6144+768
fsel:		equ	fselbas	; index of selection (<32)
pos:		equ	fsel+1	; file position after offset
ppos:		equ	pos+1	; file position before offset
offset:		equ	ppos+2	; first file displayed
curdir:		equ	offset+2 ; current directory cluster
filedata:	equ	curdir+4 ; copy of "decent" files

pos2video:		; DE = starting screen address for current position
	ld	d,$48
	ld	a,(pos)
	and	8
	jr	z,p2v2
	ld	d,$50
p2v2:	ld	a,(pos)
	sla	a
	sla	a
	sla	a
	sla	a
	sla	a
	ld	e,a
	ld	a,(pos)
	and	$10
	or	e
	ld	e,a
	ret

pos2attr:		; HL = starting attribute address for current position
	ld	h,$59
	ld	a,(pos)
	and	8
	jr	z,p2a2
	ld	h,$5a
p2a2:	ld	a,(pos)
	sla	a
	sla	a
	sla	a
	sla	a
	sla	a
	ld	l,a
	ld	a,(pos)
	and	$10
	or	l
	ld	l,a
	ret
	
clrpos:	push	bc	; clear 16-char line at (pos)
	push	de
	push	hl
	call	pos2video	; video memory
	ld	h,d
	ld	l,e
	ld	c,8
cps1:	xor	a
	ld	b,16
cps2:	ld	(hl),a
	inc	hl
	djnz	cps2
	ld	de,256-16
	add	hl,de
	dec	c
	jr	nz,cps1
	call	pos2attr	; attribute memory
	ld	a,$0E
	ld	b,16
cps3:	ld	(hl),a
	inc	hl
	djnz	cps3
	pop	hl
	pop	de
	pop	bc
	ret	

;-------------------------
; File selector
;-------------------------
filesel:
	
	in	a,($7F)	; Empty UART RX

	ld	a,$E	; Yellow over Blue
	call	cls
	; print info
	ld	de,$4000
	ld	hl,txto1
	call	vputs
	ld	de,$4000+64
	ld	hl,txto4
	call	vputs
	ld	de,$4000+128
	ld	hl,txto5
	call	vputs

	; start at root directory
	ld	de,curdir
	ld	hl,root_cluster
	ld	bc,4
	ldir

fs00:	ld	iy,fselbas	; fsel=pos=ppos=offset=0
	xor	a
	ld	(iy+fsel-fselbas),a
	ld	(iy+pos-fselbas),a
	ld	(iy+ppos-fselbas),a
	ld	(iy+ppos-fselbas+1),a
	ld	(iy+offset-fselbas),a
	ld	(iy+offset-fselbas+1),a

fs0:	; read current directory
	ld	de,filecluster
	ld	hl,curdir
	ld	bc,4
	ldir
	
	xor	a
	ld	(filesector),a
	ld	a,$ff
	ld	(filelen+3),a
	
	;------------------

fs1:	ld	ix,fatbuf	; scan directory entries
	call	rdfile	
	ld	ix,fatbuf
	ld	b,16
fs2:	ld	a,[ix+0]
	or	a
	jp	z,fs99
	cp	$E5		; deleted
	jp	z,fsskip
	ld	a,[ix+11]
	and	$30		; File or Dir
	jp	z,fsskip
	; File OK: count it
	ld	a,(iy+ppos-fselbas)
	sub	(iy+offset-fselbas)
	ld	a,(iy+ppos-fselbas+1)
	sbc	(iy+offset-fselbas+1)
	jr	z,fss1
	inc	(iy+ppos-fselbas)
	jr	nz,fsskip
	inc	(iy+ppos-fselbas+1)
	jr	fsskip
fss1:	; copy file metadata
	ld	l,(iy+pos-fselbas)
	ld	h,0
	add	hl,hl	; hl*=32
	add	hl,hl
	add	hl,hl
	add	hl,hl
	add	hl,hl
	ld	de,filedata
	add	hl,de
	ex	de,hl
	push	ix
	pop	hl
	push	bc
	ld	bc,32
	ldir
	pop	bc
	; print name, register file, register file
	call	pos2attr
	ld	a,(iy+fsel-fselbas)
	cp	(iy+pos-fselbas)
	ld	a,$0E	; Yellow over Blue
	jr	nz,fss2
	ld	a,$27	; White over Green
fss2:	push	bc
	ld	b,14	; Set line colors
fs25:	ld	(hl),a
	inc	hl
	djnz	fs25
	pop	bc
	ld	a,[ix+11] ; save file attributes
	push	af
	xor	a	; Terminate string
	ld	[ix+11],a
	ld	a,[ix+8]
	push	af	; save first Ext chararter
	xor	a	; Terminate string
	ld	[ix+8],a
	push	ix
	pop	hl
	call	pos2video
	call	vputs	; Print name
	pop	af
	ld	[ix+8],a
	inc	e
	call	vputs	; Print Extension
	pop	af
	and	$10	; Print 'D' if subdir
	ld	hl,txtfil
	jr	z,fs4
	ld	hl,txtdir
fs4:	call	vputs
	inc	(iy+pos-fselbas)
	ld	a,32
	cp	(iy+pos-fselbas)
	jr	z,fs99
fsskip:			; Next file entry
	ld	de,32
	add	ix,de
	dec	b
	jp	nz,fs2
	
	jp	fs1	; Read more sectors of directory

	;-------------- Wait for KEY action -------------

fs99:	; clear until the end of screen
	ld	a,(iy+pos-fselbas)
	push	af
fs9900:	ld	a,32
	cp	(iy+pos-fselbas)
	jr	z,fs9903
	call	clrpos
	inc	(iy+pos-fselbas)
	jr	fs9900
fs9903:	pop	af
	ld	(iy+pos-fselbas),a

	call	getkey		; Wait for key
	cp	'P'		; Right move
	jr	nz,fs991
	ld	a,(iy+fsel-fselbas)
	add	16
	cp	(iy+pos-fselbas)
	jr	nc,fs9902
	ld	(iy+fsel-fselbas),a
	jr	fs9901
fs9902:	ld	a,32
	cp	(iy+pos-fselbas)
	jr	nz,fs9901
	ld	l,(iy+offset-fselbas)
	ld	h,(iy+offset-fselbas+1)
	ld	de,16
	add	hl,de
	ld	(iy+offset-fselbas),l
	ld	(iy+offset-fselbas+1),h
fs9901:	jp	fs999
	
fs991:	cp	'O'		; Left move
	jr	nz,fs992
	ld	a,$f0
	and	(iy+fsel-fselbas)
	jr	z,fs9912
	ld	a,(iy+fsel-fselbas)
	sub	16
	ld	(iy+fsel-fselbas),a
	jr	fs9911
fs9912:	ld	a,$F0
	and	(iy+offset-fselbas)
	or	(iy+offset-fselbas+1)
	jr	z,fs9911
	ld	l,(iy+offset-fselbas)
	ld	h,(iy+offset-fselbas+1)
	ld	de,$10000-16
	add	hl,de
	ld	(iy+offset-fselbas),l
	ld	(iy+offset-fselbas+1),h
fs9911:	jp	fs999
	
fs992:	cp	'A'		; Down move
	jr	nz,fs993
	inc	(iy+fsel-fselbas)
	ld	a,(iy+fsel-fselbas)
	cp	(iy+pos-fselbas)
	jr	c,fs9921
	ld	a,(iy+pos-fselbas)
	dec	a
	ld	(iy+fsel-fselbas),a
fs9921:	jp	fs999
	
fs993:	cp	'Q'		; Up move
	jr	nz,fs994
	ld	a,(iy+fsel-fselbas)
	or	a
	jr	z,fs9931
	dec	(iy+fsel-fselbas)
fs9931:	jp	fs999
	
fs994:	cp	'B'		; Basic
	jp	z,gorom
	cp	10		; Enter or Space
	jr	z,fs9941
	cp	32
	jr	z,fs9941
	jp	nz,fs998
fs9941:	ld	l,(iy+fsel-fselbas)
	ld	h,0
	add	hl,hl	; hl*=32
	add	hl,hl
	add	hl,hl
	add	hl,hl
	add	hl,hl
	ld	de,filedata
	add	hl,de
	push	hl
	pop	ix
	ld	a,$10	; Subdir?
	and	(ix+11)
	jr	z,fs995
	ld	l,(ix+26)
	ld	h,(ix+27)
	ld	e,(ix+20)
	ld	d,(ix+21)
	ld	a,l	; cluster==0 => root
	or	h
	or	e
	or	d
	jp	z,filesel
	ld	ix,curdir
	ld	(ix+0),l
	ld	(ix+1),h
	ld	(ix+2),e
	ld	(ix+3),d
	ld	a,$ff
	ld	(filelen),a
	jp	fs00

fs995:	ld	a,'S'	; SNA
	cp	(ix+8)
	jr	nz,fs996
	ld	a,'N'
	cp	(ix+9)
	jr	nz,fs996
	ld	a,'A'
	cp	(ix+10)
	jr	nz,fs996
	ld	l,(ix+26)
	ld	h,(ix+27)
	ld	e,(ix+20)
	ld	d,(ix+21)
	ld	(filecluster),hl
	ld	(filecluster+2),de
	ld	l,(ix+28)
	ld	h,(ix+29)
	ld	e,(ix+30)
	ld	d,(ix+31)
	ld	(filelen),hl
	ld	(filelen+2),de
	xor	a
	ld	(filesector),a
	jp	loadSNA
	
fs996:	ld	a,'Z'	; Z80
	cp	(ix+8)
	jr	nz,fs997
	ld	a,'8'
	cp	(ix+9)
	jr	nz,fs997
	ld	a,'0'
	cp	(ix+10)
	jr	nz,fs997
	ld	l,(ix+26)
	ld	h,(ix+27)
	ld	e,(ix+20)
	ld	d,(ix+21)
	ld	(filecluster),hl
	ld	(filecluster+2),de
	ld	l,(ix+28)
	ld	h,(ix+29)
	ld	e,(ix+30)
	ld	d,(ix+31)
	ld	(filelen),hl
	ld	(filelen+2),de
	xor	a
	ld	(filesector),a
	ld	bc,0
	ld	(bufix),bc
	ld	(bufmax),bc
	jp	loadZ80

fs997:	ld	a,'T'	; TAP file
	cp	(ix+8)
	jp	nz,fs998
	ld	a,'A'
	cp	(ix+9)
	jp	nz,fs998
	ld	a,'P'
	cp	(ix+10)
	jp	nz,fs998
	ld	l,(ix+26)
	ld	h,(ix+27)
	ld	e,(ix+20)
	ld	d,(ix+21)
	ld	(filecluster),hl
	ld	(filecluster+2),de
	ld	l,(ix+28)
	ld	h,(ix+29)
	ld	e,(ix+30)
	ld	d,(ix+31)
	ld	(filelen),hl
	ld	(filelen+2),de
	xor	a
	ld	(filesector),a
	ld	a,4
	call	cls2
	in	a,($FF)
	and	$F9		; Page 2 to $8000 - $FFFF
	out	($FF),a
	ld	de,$8001	; Clear page
	ld	hl,$8000
	ld	bc,$7FFF
	xor	a
	ld	(hl),a
	ldir
	ld	de,$8000
ltp1:	ld	a,$80
	and	d
	jr	nz,ltp2
	in	a,($FF)
	or	2		; Page 3 to $8000 - $FFFF
	out	($FF),a
	ld	de,$8001	; Clear page
	ld	hl,$8000
	ld	bc,$7FFF
	xor	a
	ld	(hl),a
	ldir
	ld	de,$8000
ltp2:	ld	ix,fatbuf
	call	rdfile
	push	bc
	ld	hl,fatbuf
	ldir
	pop	bc
	ld	a,2
	and	b
	jr	nz,ltp1
	in	a,($FF)
	or	6		; Page 1 to $8000 - $FFFF
	out	($FF),a
	
	ld	hl,txtotp
	ld	de,$5000
	call	vputs
	ld	bc,50
	call	delay
	ld	a,$E
	call	cls2
	jp	fs999
	
fs998:	cp	'H'		; Hex dump (debug)
	jr	nz,fs999
	in	a,($FF)
	and	$F9
	out	($FF),a

	ld	hl,$8000
	call	hdump2
	ld	a,10
	call	putch

	in	a,($FF)
	or	6
	out	($FF),a
	
fs999:	xor	a
	ld	(iy+pos-fselbas),a
	ld	(iy+ppos-fselbas),a
	ld	(iy+ppos-fselbas+1),a
	jp	fs0

;--------------------------------------
; delay loop
; args: BC: number of loops
; modiffies: A, BC (returns as 0)

delay:	push	bc
	ld	bc,0
dl1:	dec	bc
	ld	a,b
	or	c
	jr	nz,dl1
	pop	bc
	dec	bc
	ld	a,b
	or	c
	jr	nz,delay
	ret

;--------------------------------------
; Clear Screen
; args: A: text attributes
; modiffies: A, BC, DE, HL

cls:	ld	hl,$4000
	ld	b,0
	ld	(hl),b
	ld	de,$4001
	ld	bc,6144-1
	ldir
	inc	hl
	inc	de
	ld	(hl),a
	ld	bc,768-1
	ldir
	ret
	
; Clear only the bottom 2/3 of the screen
cls2:	ld	hl,$4800
	ld	b,0
	ld	(hl),b
	ld	de,$4801
	ld	bc,4096-1
	ldir
	ld	hl,$4000+6144+256
	ld	de,$4000+6144+256+1
	ld	(hl),a
	ld	bc,512-1
	ldir
	ret


;--------------------------------------
;--------------------------------------
; Load .Z80 file
;--------------------------------------
;--------------------------------------

loadZ80:
	; Header with register values
	ld	b,30
	ld	hl,varbas
lz1:	call	fstream
	ld	(hl),a
	inc	hl
	djnz	lz1	

	; Versions 2 & 3 have PC=$0000 and an extra header
	ld	bc,(varbas+6)
	ld	a,b
	or	c
	jr	nz,z80v1
	call	fstream	; extra lenght
	ld	b,a
	call	fstream
	call	fstream ; PC_L
	ld	(varbas+6),a
	call	fstream ; PC_H
	ld	(varbas+7),a
	dec	b
	dec	b	
lzv23:	call	fstream	; get rid of extra bytes
	djnz	lzv23
	ex	af,af'	; number of pages in A'
	ld	a,3
	ex	af,af'
lzv21:	call	fstream ; page len (16-bit, ignored)
	call	fstream
	call	fstream ; page number
	ld	h,$80	; page 4 @ $8000
	cp	4
	jr	z,lv22
	ld	h,$C0
	cp	5	; page 5 @ $C000
	jr	z,lv22
	ld	h,$40	; others @ $4000
lv22:	ld	l,0
	ld	bc,$4000 ; page size
	call	unzip
	ex	af,af'
	ld	b,a
	dec	a
	ex	af,af'
	dec	b	
	jr	nz,lzv21
	jp	goz80
	
	; Version 1
z80v1:	ld	a,(varbas+12)
	and	$20
	jr	nz,lz21
	
lz20:	call	fstream	; Without compression
	ld	(hl),a
	inc	hl
	ld	a,h
	or	l
	jr	nz,lz20
	jr	lz9
	
lz21:	; Compressed data
	ld	hl,$4000
	ld	bc,$c000
	call	unzip
lz9:	jp	goz80

unzip:	; decompress data block
	; $ED $ED nn nn => nn bytes with nn value
	; else data
lz2:	call	fstream
	cp	$ED
	jr	nz,lz3
	call	fstream	; <esc>
	cp	$ED	
	jr	nz,lz5
			; <esc>, <esc>: follows COUNT and DATA
	call	fstream	; COUNT
	ld	d,a
	call	fstream	; DATA
	ld	e,a
lz4:	ld	(hl),e
	inc	hl
	dec	bc
	ld	a,b
	or	c
	ret	z
	dec	d
	jr	nz,lz4
	jr	lz2
	
lz5:	ld	d,a	; <esc>, /<esc> : save data
	ld	e,$ED
	ld	(hl),e
	inc	hl
	dec	bc
	ld	a,b
	or	c
	ret	z
	;ld	(hl),d
	;inc	hl
	;dec	bc
	;ld	a,b
	;or	c
	;ret	z
	;jr	lz2
	ld	a,d
	
lz3:	ld	(hl),a	; simple data
	inc	hl
	dec	bc
	ld	a,b
	or	c
	jr	nz,lz2
	ret
;--------------------------------------
; boot Z80
;--------------------------------------
	; Clear ROM area
goz80:	ld	de,fatbas+1
	ld	hl,fatbas
	ld	bc,$3d00-fatbas-1
	ld	a,$ff
	ld	(hl),a
	ldir

	; Load registers and jump to code
	; Copy trampoline to RAM
	ld	de,fatbas
	ld	hl,bootZ80
	ld	bc,ebootZ80-bootZ80
	ldir
	ex	de,hl		; Adjust JMP addr
	dec	hl
	ld	a,(varbas+7)	; PCH
	ld	(hl),a
	dec	hl
	ld	a,(varbas+6)	; PCL
	ld	(hl),a
	ld	a,(varbas+27)	; IFF2
	or	a
	jr	nz,lz91
	dec	hl		; skip JP
	dec	hl		; Turn EI into NOP
	ld	(hl),a
lz91:	
	im 	1		; Adjust Interrupt mode
	ld	a,(varbas+29)
	and	2
	jr	z,lz92
	im	2
lz92:	ld	a,(varbas+12)	; Border Color
	rra	
	and	7
	out	($FE),a
	ld	a,(varbas+10)	; I reg
	ld	i,a
				; register R is ommited
	exx			; Alternate register pairs
	ld	bc,(varbas+15)
	ld	de,(varbas+17)
	ld	hl,(varbas+19)
	exx
	ld	de,(varbas+13)	; DE
	ld	hl,(varbas+4)	; HL
	ld	bc,(varbas+23)	; IY
	push	bc
	pop	iy
	ld	bc,(varbas+25)	; IX
	push	bc
	pop	ix
	ex	af,af'		; Alternate Accumulator and Flags
	ld	a,(varbas+21)	; reverse low-high order
	ld	b,a
	ld	a,(varbas+22)
	ld	c,a
	push	bc
	pop	af
	ex	af,af'
	ld	a,(varbas+0)	; Accumulator and Flags
	ld	b,a		; reverse low-high order
	ld	a,(varbas+1)
	ld	c,a
	push	bc
	pop	af
	ld	sp,(varbas+8)	; Stack pointer

	jp	fatbas		; Jump to trampoline in RAM (later ROM)
	
bootZ80:	; Trampoline code (copied to RAM)
	ld	bc,$37FF	; BOOT off, pag 11
	out	(c),b
	ld	bc,(varbas+2)	; restore BC
	ei			; replaced with NOP if no interrupts
	jp	0		; Address replaced with PC content
ebootZ80:

;--------------------------------------------------------------------
;--------------------------------------------------------------------
;	Keyboard Scanning
;--------------------------------------------------------------------
;--------------------------------------------------------------------

ktab:	db	"cZXCV"		; 'c' => Caps Shift
	db	"ASDFG"
	db	"QWERT"
	db	"12345"
	db	"09876"
	db	"POIUY"
	db	10,"LKJH"	; ENTER as LF
	db	" sMNB",0	; 's' => Symbol Shift, 0 => No key

; single keyboard scan
kscan:	push	bc
	push	de
	push	hl
	ld	d,$FE	; // First row
	ld	e,8
	ld	hl,ktab
ks1:	ld	a,d
	in	a,($FE)
	scf
	rl	d
	ld	b,5
ks2:	rra
	jr	nc,ks9
	inc	hl
	djnz	ks2
	dec	e
	jr	nz,ks1
ks9:	ld	a,(hl)
	pop	hl
	pop	de
	pop	bc
	ret

; Wait for key pressed (with repetition)
getkey:	ld	d,2
gk0:	call	delay2
	call	kscan
	or	a
	jr	z,gk1
	dec	d	; still pressed
	jr	nz,gk0
	ret	
	; wait for key press
gk1:	call	delay2	
	in	a,($BF)	; or for UART RX
	and	1
	jr	z,gk2
	in	a,($7F)
	ret
gk2:	call	kscan
	or	a
	jr	z,gk1
	ret

; delay for key repetition
delay2:	ld	bc,$4000
dla1:	dec	bc
	ld	a,b
	or	c
	jr	nz,dla1
	ret
