//--------------------------------------------------------------------
// Sistema con CPU, RAM y UART
//--------------------------------------------------------------------

module SYSTEM (
	input clk,
	input reset,

	output xwe,		// Bus externo: WE
	output xoe,		// Bus externo: OE
	output [15:0]xa,// Bus externo: direcciones
	output [15:0]xdo,	// Bus externo: output
	input  [15:0]xdi,	// Bus externo: input
	output xbhe,	// Bus externo: Byte alto
	output xble,	// Bus externo: byte bajo

	output txd,
	input rxd,

	output hsyn,
	output vsyn,
	output [3:0]r,
	output [3:0]g,
	output [3:0]b,

	input  [7:0]pinin,
	output [7:0]pinout,

	output	pwmout,		// AUDIO
`ifdef NES
	output	gclk,
	output	glat,
	input	gin,
`endif
	input	kclk,
	input	kdat,

	output [1:0]debug	
);

wire res;

///////////// Divisor de reloj //////////////
wire fast;
assign	fast=bootreg[3] | play;
reg [2:0]divi=0;
always @(posedge clk) divi<=(divi[2]&divi[1])?0 : divi+1;
wire cclk;
assign cclk=fast ? clk : divi[2];

/////////////////////////////// CPU ////////////////////////////////
wire reset_n,wait_n,int_n,nmi_n,busrq_n,cen;
wire [7:0]di;
wire [7:0]dout;
wire [15:0]A;
wire m1_n,mreq_n,iorq_n,rd_n,wr_n,rfsh_n,halt_n,busak_n;

assign res=reset|resetk;

assign reset_n=~res;
assign wait_n=1'b1;

assign nmi_n=1'b1;
assign busrq_n=1'b1;
assign cen=1'b1;

tv80s Z80uno(//AUTOARG
  // Outputs
  m1_n, mreq_n, iorq_n, rd_n, wr_n, rfsh_n, halt_n, busak_n, A, dout,
  // Inputs
  reset_n, cclk, wait_n, int_n, nmi_n, busrq_n, di, cen
  );

wire boot;		// modo especial de arranque
assign boot=bootreg[3];

// Memoria externa

wire [16:0]mxa;	// Ext. mem. address
wire [7:0]mxdi;

// Paging
assign mxa=mreq_n ? {1'b1,tpaddr}: 
		   {(A[15]?{~bootreg[2],bootreg[1]}:2'b00),A[14:0]};

assign xbhe=~mxa[0];
assign xble=mxa[0];
assign xoe= ((~wr_n)&(~mreq_n)); //~((~mreq_n)&(~rd_n));
wire noromw;
assign noromw=(~A[15])&(~A[14])&(~boot);
assign xwe=~((~mreq_n)&(~wr_n)&(~noromw));
assign xa=mxa[16:1];
assign xdo={dout,dout};
assign mxdi=mxa[0]?xdi[15:8]:xdi[7:0];

//////// Señales de control adicionales
//wire iowr;
//assign iowr =(~iorq_n)&(~wr_n);
wire [7:0]dorom;	// salida de memorias
wire [7:0]doio;		// salida de periféricos
assign di=(~mreq_n)?(romcs?dorom:mxdi):doio;

///// Memoria interna
wire romcs;
//assign romcs=(~mreq_n)&(~A[15])&(~A[14])&(~A[13])&(~A[12])&(~A[11])&(~A[10])&(~A[9])&boot;
assign romcs=(~mreq_n)&(~A[15])&(~A[14])&(~A[13])&boot;

genrom #(
`ifdef SIMULATION
	.INITFILE("rom.hex"),
`else
	.INITFILE("rand.hex"),
`endif
	.AW(13),
	.DW(8)
	) rom0 (.clk(~cclk),.addr(A[12:0]),.data_out(dorom));

////// Video /////
wire [3:0]igrb;
wire vwr;
assign vwr=(~mreq_n)&(~wr_n)&(~A[15])&A[14]&(~A[13]);

video video0 (.clk(clk),.clkw(clk),.di(dout),.aw(A[12:0]),.we(vwr),
			  .hsyn(hsyn),.vsyn(vsyn),.video(igrb),.border(ulareg[2:0]) );

assign b[3]=igrb[0];
assign b[2]=igrb[0];
assign b[1]=igrb[0]&igrb[3];
assign b[0]=igrb[0]&igrb[3];
assign r[3]=igrb[1];
assign r[2]=igrb[1];
assign r[1]=igrb[1]&igrb[3];
assign r[0]=igrb[1]&igrb[3];
assign g[3]=igrb[2];
assign g[2]=igrb[2];
assign g[1]=igrb[2]&igrb[3];
assign g[0]=igrb[2]&igrb[3];


//////////////////// PERIFERICOS /////////////////////

wire csFE;	// selección ULA
wire csFF;	// selección bootReg
wire cs7F;	// selección UART (datos)
wire csDF;	// selección joystick

assign csFF = (~iorq_n)&(A[7:0]==8'hFF)&boot;
assign csFE = (~iorq_n)&(~A[0]);
assign csDF = (~iorq_n)&(~A[5]);
assign cs7F = (~iorq_n)&(~A[7])&boot;
//assign csBF = (~iorq_n)&(A[7:0]==8'hBF)&boot;

reg [4:0]ulareg=0;
reg [7:0]bootreg=8'hff;

always @(posedge cclk) begin
	if (csFE&(~wr_n)) ulareg<=dout;
end
always @(posedge cclk or posedge res) begin
	if (res) bootreg<=8'hff;
		else if (csFF&(~wr_n)) bootreg<=dout;
end

// AUDIO
reg [5:0]aucnt=0;
always @(posedge cclk) aucnt<=aucnt+1;
assign pwmout=(aucnt[5]&aucnt[4]&aucnt[3])?ulareg[3]:ulareg[4];
//assign pwmout=ulareg[3]^ulareg[4];

// SPI
assign pinout[7:4]=bootreg[7:4];

///// UART
wire [7:0]uartdo;
wire thre,dv,fe,ove;
UART_CORE #(.DIVISOR(217))
	UART1(	.txd(txd), .tend(tend), .thre(thre),.d(dout),.wr(cs7F&(~wr_n)),
		.q(uartdo), .dv(dv), .fe(fe), .ove(ove),
		.rxd(rxd), .rd(cs7F&(~rd_n)), .clk(cclk) );

`ifdef NES
///// NES GAME Controller (shift reg. 74HC4021)
reg [3:0]gc=0;
always @(posedge hsyn) gc<=gc+1;
assign glat=(~gc[3])&(~gc[2])&(~gc[1])&(~gc[0]);
assign gclk=~gc[0];
reg [7:0]gsh=0;
reg [7:0]gdat;
always @(negedge gclk) gsh<={gsh[6:0],gin};
always @(posedge glat) gdat<=~gsh;
`endif


////// Multiplexado de datos de entrada de periféricos
assign doio = csFF ? {bootreg[7:1],pinin[7]} : 
			(csFE ? {1'b1,casdat,1'b1,krow} : 	// ULA
`ifdef NES
			(csDF ? {gdat[5:4],gdat[6],gdat[7],gdat[3:0]} : // NES
`else			
			(csDF ? {~pinin[6],1'b0,~pinin[5:0]} : // Joystick Kempston
`endif
			(cs7F ? uartdo : {3'b000,ove,fe,tend,thre,dv})));	// UART

////// Teclado //////
wire [4:0]krow;
wire resetk,tapek;

keybps2 keycon( .clk(cclk), .kclk(kclk), .kdat(kdat),.a(A[15:8]),.row(krow),.resetk(resetk), .tapek(tapek));

///// Interrupción (59.5Hz)

reg irq=0;
reg ovs=0;

always @(posedge cclk) begin
	ovs<=vsyn;
	if (~(iorq_n | m1_n)) irq<=0;
	else if ((~vsyn)&ovs) irq<=1;
end
assign int_n=~irq;

//////// TAPE Player ///////

wire [15:0]tpaddr;
reg [7:0]tpdat;
wire casdat;
wire play;

always @(posedge cclk) if (mreq_n) tpdat<=mxdi;

//--- divisor clk (1/70)

reg [6:0]tpdiv=0;
wire tpclk;
always @(posedge cclk) tpdiv<=(tpdiv==69)? 0 : tpdiv+1;
assign tpclk=tpdiv[6];

reg start=0;
always @(posedge cclk) begin
	if (tapek) start<=1;
	else if (play) start<=0;
end

tape tape_player( .clk(tpclk), .resb(reset_n), .di(tpdat), .start(start), 
				  .addr(tpaddr), .Q(casdat), .play(play), .debug(debug) );

endmodule


//----------------------------------------------------------------------------
//-- Memoria ROM genérica
//----------------------------------------------------------------------------

module genrom (        //-- Puertos
    input clk,                      //-- Señal de reloj global
    input wire [AW-1: 0] addr,      //-- Direcciones
    output reg [DW-1: 0] data_out   //-- Dato a escribir
);
//-- Parametro: Nombre del fichero con el contenido de la RAM
parameter INITFILE = "rand.hex";
parameter AW = 13;
parameter DW = 8;

//-- Calcular el numero de posiciones totales de memoria
//localparam NPOS = 2 ** AW;
//localparam NPOS = 1<<AW;

  //-- Memoria
  reg [DW-1: 0] rom [0: (1<<AW)-1];

  //-- Escritura en la memoria
  always @(posedge clk) begin
    data_out <= rom[addr];
  end

//-- Cargar en la memoria el fichero INITFILE
//-- Los valores deben estan dados en hexadecimal
initial begin
  $readmemh(INITFILE, rom);
end

endmodule

