//--------------------------------------------------------------------
//--------------------------------------------------------------------

module tape (
	input clk,
	input resb,
	input start,
	input [7:0]di,
	output Q,
	output play,
	output [15:0]addr,
	output [1:0]debug
);

/////////////////////////////////
// Generador de pulsos
/////////////////////////////////

reg [4:0]pcount=0;

// Inputs
wire pilot;
wire sync;
wire dat;

assign pilot=(st==PILOT)|(st==GAP);
assign sync=(st==SYNC);
assign dat=shreg[7];


// Output
reg toggle=0;
//reg otog=0;
//wire rising;

wire [4:0]recarga;
assign recarga= pilot ? 30 : (sync ? 9 : ( dat ? 23 : 11));
always @(posedge clk) begin
	if (pcount==0) begin
		pcount<= recarga;
		toggle<=~toggle;
	end else pcount<=pcount-1;
	//otog<=toggle;
end
//assign rising = toggle & (~otog);

/////////////////////////////////
// bit shift
/////////////////////////////////

reg	[7:0]shreg=65;
wire shift;
wire shload;
assign shload=(st==SYNC) | ((bcount[2:0]==3'b000)&(st==DATA));

always @(posedge toggle) begin
	if (shload) shreg<=di;
	else shreg<={shreg[6:0],1'b0};
end

/////////////////////////////////
// bit counter
/////////////////////////////////
reg [15:0]bytecount=0;
reg [18:0]bcount=3*8;
wire bcload;
wire bcpres;
wire endblock;
wire einc;
assign bcload=(st==SYNC);
assign bcpres=(st==RDH) | ((st==DATA)&endblock);

always @(posedge toggle) begin
	if (bcpres) bcount<=4095; //2047;
	else begin
		if (bcload) bcount<={bytecount,3'b111};
		else bcount<=bcount-1;
	end
	if (st==RDL) bytecount[ 7:0]<=di;
	if (st==RDH) bytecount[15:8]<=di;
end
assign endblock=(bcount[18:3]==0);
assign einc=~(bcount[18:3]==1);
/////////////////////////////////
// Memory Address
/////////////////////////////////
reg [15:0]addr=0;
wire inc;
wire addres;
assign inc= einc&((st==RDL)|(st==RDH)| shload); 

always @(posedge toggle) begin
	if (st==IDLE) addr<=0;
	else if (inc) addr<=addr+1;
end

/////////////////////////////////
// State Machine
/////////////////////////////////
// Asignación de estados
parameter IDLE=		0;
parameter RDL=		1;
parameter RDH=		2;
parameter PILOT=	3;
parameter SYNC=		4;
parameter DATA=		5;
parameter GAP=		6;
parameter UNUSED=	7;

reg  [2:0]st=0; // Estado actual
// Lógica del estado siguiente
always @(posedge toggle or negedge resb)
begin
	if (!resb) st<=IDLE;
	else 
		case (st)
		IDLE:	if (start) st<=RDL;
		RDL:	st<=RDH;
		RDH:	if ((bytecount[7:0]==0)&(di==0)) st<=IDLE; else st<=PILOT;
		PILOT:	if (endblock) st<=SYNC;
		SYNC:	st<=DATA;
		DATA:	if (endblock) st<=GAP;
		GAP:	if (endblock) st<=RDL;
		default:	st<=3'bxxx;
		endcase
end

reg enq=0;
always @(posedge toggle) begin
	enq<=((st==PILOT)|(st==SYNC)|(st==DATA));
end

assign Q=toggle&enq;
assign play=(st!=IDLE);

assign debug[0]=addr[0];
assign debug[1]=Q;

endmodule

