
//---------------------------------------------------------------------------------- 
//		Core de UART
//	- Doble función de transmisor y receptor
//----------------------------------------------------------------------------------

module UART_CORE(output txd, output tend, output thre, input [7:0]d, input wr,
			output [7:0]q, output dv, output fe, output ove, input rxd, input rd, input clk );

parameter DIVISOR= 6;
localparam NDIV = $clog2(DIVISOR);

///////////////////////////////////////
// Transmisor
reg [7:0]buftx;		// Buffer anterior al registro de desplazamiento
reg bufdata=0;		// FF de estado 1: buffer vacío, 0 buffer con dato
reg [8:0]shtx=9'h1ff;		// Reg. desplazamiento de 9 bits (para incluir bit de start)
reg [3:0]cntbit;	// Contador de bits transmitidos
reg rdy=1;			// Estado del resgistro de desplazamiento (1==idle)

// Divisor de TX (1/16)

reg [NDIV-1:0] divtx=0;
wire clko;			// pulsos de 1 ciclo de salida
assign clko = (divtx==0);
always @ (posedge clk) divtx <= clko ? DIVISOR-1: divtx-1;

always @(posedge clk)
begin
	// buffer Tx
	if (wr) begin
`ifdef SIMULATION
		$write ("%c",d&255); $fflush ( );
`endif
		buftx<=d;
		bufdata<=1;
	end
	if (clko) begin
		if(rdy&bufdata) begin
			rdy<=0;
			bufdata<=0;
			shtx<={buftx[7:0],1'b0};
			cntbit<=0;
		end
		if(~rdy) begin
			shtx<={1'b1,shtx[8:1]};
			cntbit<=cntbit+1;
			if (cntbit[3]&cntbit[1]) #1 rdy<=1;
		end
	end

end

assign txd = shtx[0];
assign thre = ~bufdata;
assign tend = (~bufdata)&rdy;

////////////////////////////////////////
// Receptor

/// Sincro
reg [1:0]rrxd=2'b11;	// RXD registrada dos veces
wire resinc;			// activa si cambio en RXD (resincroniza divisor)
wire fedge;				// activa si flanco de bajada en RXD (para start)
always @(posedge clk) #1 rrxd<={rrxd[0],rxd};
assign resinc = rrxd[0]^rrxd[1];
assign fedge = (~rrxd[0])&rrxd[1];

/// Divisor
// Genera un pulso en mitad de la cuenta (centro de bit)
// se reinicia con resinc
reg [NDIV-1:0] divrx=0;
wire clki;		// Pulso de 1 ciclo de salida
wire clki0;		// recarga de contador
assign clki = (divrx==(DIVISOR/2+1));
assign clki0= (divrx==0);

always @ (posedge clk) divrx <= (resinc|clki0) ? DIVISOR-1: divrx-1;

reg dv=0;				// Dato válido si 1
reg ove=0;				// Overrun
reg [8:0]shrx;			// Reg. desplazamiento entrada (9 bits para stop)
reg [7:0]q=8'hff;		// Buffer RX
reg stopb=1'b1;			// Bit de stop recibido
reg [3:0]cbrx=4'b1111;	// Contador de bits / estado (1111== idle)
wire rxst;				// Guardar shr en buffer si flanco subida
assign rxst=(cbrx==4'b1111);
reg rxst0;				// Para detectar flanco

always @(posedge clk)
begin
	rxst0<=rxst;
	if (rxst&fedge) cbrx<=9;
	if (clki & (~rxst)) begin
		shrx<= {rrxd[0],shrx[8:1]};
		cbrx<=cbrx-1;
	end
	if (rxst & (~rxst0)) begin 
		{stopb,q}<=shrx;
		dv<=1;
		ove<=dv;
	end

	if (rd) begin
		dv<=0;
		ove<=0;
	end
end

assign fe=~stopb;

endmodule

